/* ITKA203 Käyttöjärjestelmät kesä 2011
 * Malliharjoitustyö
 *
 * Tässä on C-kielinen ohjelma, joka (jalavan x86-64 -tietokoneessa
 * gcc:llä käännettynä) tekee seuraavaa:
 *
 *      - summaa 64-bittisiä kokonaislukuja sisältävän taulukon
 *        alkiot taulukon tietyltä väliltä, ja summauksen yhteydessä
 *        asettaa alkiot nolliksi. Kohdistamalla indeksivälille 1..3
 *        taulukosta {45, -4, 56, -57, -4, 9} pitäisi tulla 
 *        {45, 0, 0, 0, -4, 45} ja summaksi palautua -5.
 *
 * Ohjelmasta ei varmaan ole hyötyä muuhun kuin sen pedagogiseen
 * tarkoitukseen.
 *
 * Kaikki muuttujat ovat tyypiltään ``long long int`` jotta ne jalavan
 * gcc:llä tulisivat kokonaisen 64-bittisen rekisterin kokoisiksi.
 * Syy on pedagoginen (helpompi tulkita konekielistä käännöstä ja
 * debuggerin tulosteita).
 */

#include<stdio.h>

/**
 * Sattuneesta syystä varsinaisesta esimerkistä tulee mielekkäämpi,
 * kun siinä tehdään välissä yksi kutsu "turhaan" aliohjelmaan, jossa
 * ei tehdä mitään.
 * 
 * (Kutsu nimittäin varmistaa, että itse tutkittavana oleva aliohjelma
 * toimii "kunnon kansalaisen" tavoin, eikä suorituskykyoptimoidun
 * "eliitin", joka tietää olevansa aina kutsupinon päällimäinen;
 * lisätietoa saa googlettamalla "amd64 red zone"; tällaiset ABI:n
 * detaljit eivät oikeastaan kuulu kurssin sisältöön, joten
 * hyväksytään nyt että tässä on tällainen turhalta vaikuttava
 * aliohjelma, joka ei tee mitään muuta kuin palaa heti takaisin
 * kutsuvaan ohjelmaan);
 */
void turha_kutsu(){
}

/** 
 * Aliohjelma, joka manipuloi taulukkoa. Parametrina on C:n taulukko,
 * eli käytännössä ensimmäisen alkion muistiosoite, sekä operaation
 * alku- ja loppuindeksit.
 */
long long int summaa_ja_nollaa_valilta(long long int taulu[], 
                             long long int indA,
                             long long int indB){
  long long int i;
  long long int summa;
  turha_kutsu();
  summa=0;
  for(i=indA; i<=indB; i++){
    summa += taulu[i];
    taulu[i] = 0;
  }
  return summa;
}

/** Testailua varten tein pienen tulostuskoodin. */
void tulosta_taulukko(long long int taulu[], long long int koko){
  int i;

  for(i=0; i<koko; i++){
    printf("%lld ", taulu[i]);
  }
  printf("\n");
}

/**
 * Pääohjelma (joka C:ssä on käytännössä vain sopivasti nimetty
 * aliohjelma).
 */
int main(int argc, char *argv[]){
  long long int testitaulu[] = {-1, 7, 3, -16, 17, 8};
  long long int alku = 2;
  long long int loppu = 5;
  long long int s = 0;

#ifndef HILJAA
  tulosta_taulukko(testitaulu, 6);
#endif

  s = summaa_ja_nollaa_valilta(testitaulu, alku, loppu);

#ifndef HILJAA
  printf("Summa indeksien %lld ja %lld väliltä oli %lld.\n", 
         alku, loppu, s);
  printf("Taulukko lopuksi, väli nollattuna, on:\n");
  tulosta_taulukko(testitaulu, 6);
#endif

  return 0;
}
