/* ITKA203 Käyttöjärjestelmät kevät 2014, kevät 2015, kevät 2016.
 *
 * "Malliharjoitustyö" tutkittavaksi automaattisesti lokiin
 * kirjautuvan jäljen perusteella.
 *
 * Tämä C-kielinen ohjelma tekee seuraavaa:
 *
 *      - kääntää 64-bittisiä kokonaislukuja sisältävän taulukon
 *        alkiot käänteiseen järjestykseen, esimerkiksi taulukosta
 *        {45, -4, 56, 57, -4, 9} pitäisi tulla {9, -4, 57, 56, -4,
 *        45}. Operaatio tehdään "paikallaan" eli vaihdetaan keskenään
 *        ensimmäinen ja viimeinen, toinen ja toiseksi viimeinen, ja
 *        niin edelleen.
 *
 *      - ilmoittaa paluuarvona, montako kertaa vaihdettiin sama luku
 *        itsensä kanssa; koska em. esimerkissä -4 oli sekä toinen
 *        että toiseksi viimeinen alkio, pitäisi paluuarvoksi tulla
 *        ykkönen.
 *
 * Ohjelmasta ei varmaan ole hyötyä muuhun kuin sen pedagogiseen
 * tarkoitukseen: Voimme debuggeria käyttäen nähdä, kuinka
 * aliohjelmien kutsuminen, parametrien välitys, paikallisten
 * muuttujien käyttö sekä paluuarvon palauttaminen toimii konekielen
 * tasolla.
 *
 * Kaikki muuttujat ovat tyypiltään C99:n yhteensopivuuskirjaston
 * nimeämällä tavalla ``int64_t``, jotta ne tulisivat x86-64:n
 * kokonaisen 64-bittisen rekisterin kokoisiksi. Syy myös tähän on
 * pedagoginen: on helpompi tulkita konekielistä käännöstä ja pinon
 * sisältöä, kun sekä muistiosoitteet että käsiteltävät datat ovat
 * kaikki tasan 8 tavun mittaisia.
 *
 * Käytännössä 64-bittiset luvut ovat voivat olla moneen tarpeeseen
 * tarpeettoman isoja, mutta esimerkiksi tavutaulukon tai -tiedoston
 * indeksointi 32 bitillä rajaa käsiteltävän datan määräksi noin neljä
 * gigatavua, mikä ei nykyään oikein riitä. Kellonajan ilmoittaminen
 * sekunteina keskiyöstä 1.1.1970 alkaen epäonnistuu 32-bittisellä
 * luvulla aamuyöstä tiistaina 19.1.2038
 * jne.. (http://en.wikipedia.org/wiki/Year_2038_problem) Ohjelmoijan
 * täytyy varmistaa, että ohjelmointialustan lukujen esitysmuodot
 * riittävät sovelluksen nykyisiin ja tuleviin tarpeisiin.
 */

#include<stdint.h>

#ifdef TEE_TULOSTEET
#include<stdio.h>
#include<inttypes.h>
#endif

/**
 * Sattuneesta syystä varsinaisesta esimerkistä tulee mielekkäämpi,
 * kun siinä tehdään välissä yksi kutsu "turhaan" aliohjelmaan, jossa
 * ei tehdä mitään: Kutsu varmistaa, että demotehtävässä tutkittavana
 * oleva aliohjelma toimii "normaalin kansalaisen" tavoin eikä
 * suorituskykyoptimoidun "eliitin", joka tietää olevansa aina
 * kutsupinon päällimäinen. Lisätietoa saa halutessaan googlettamalla
 * "amd64 red zone"; tällaiset ABI:n detaljit eivät oikeastaan kuulu
 * kurssin sisältöön, joten hyväksytään nyt, että tässä on tällainen
 * turhalta vaikuttava aliohjelma, joka ei tee mitään muuta kuin palaa
 * heti takaisin kutsuvaan ohjelmaan.
 */
void sakkokierros(){
}

/** 
 * Aliohjelma, joka manipuloi taulukkoa. Parametrina on C:n taulukko,
 * eli käytännössä ensimmäisen alkion muistiosoite, sekä kokonaisluku,
 * joka tulkitaan taulukon kooksi (eli alkioiden lukumääräksi).
 */
int64_t kaanna_taulukko(int64_t taulu[], int64_t koko){
  int64_t ioikea, ivasen;
  int64_t lkm_samoja = 0;
  int64_t tmp;
  ivasen=0;
  for(ioikea=koko-1; ioikea>ivasen; ioikea--){
    if (taulu[ivasen] == taulu[ioikea]){
      lkm_samoja++;
    } else {
      /* vaihto */
      tmp = taulu[ivasen];
      taulu[ivasen] = taulu[ioikea];
      taulu[ioikea] = tmp;
    }
    ivasen++;
  }
  sakkokierros();
  return lkm_samoja;
}


#ifdef TEE_TULOSTEET
/** Testailua varten tein pienen tulostuskoodin. */
void tulosta_taulukko(int64_t taulu[], int64_t koko){
  int i;
  printf("64-bittisiä lukuja %" PRId64 " kpl alkaen osoitteesta %p:\n", koko, taulu);
  for(i=0; i<koko; i++){
    printf("%" PRId64 " ", taulu[i]);
  }
  printf("\n");
}
#endif

/**
 * Pääohjelma.
 */
int main(int argc, char *argv[]){
  // Kivat lukuarvot, joita on visuaalisesti helppo seurata debuggerissa:
  // Yksi negatiivinenkin luku siellä on; voi samalla todeta, miten -1
  // koodautuu etumerkillisten lukujen 2-komplementtiesityksessä.
  int64_t testitaulu[] = {0x1000, 0x2000, 0x3000, -1, 0x2000, 0x6000};
  int64_t koko = 6;
  int64_t samoja;

#ifdef TEE_TULOSTEET
  tulosta_taulukko(testitaulu, koko);
#endif

  samoja = kaanna_taulukko(testitaulu, koko);

#ifdef TEE_TULOSTEET
  tulosta_taulukko(testitaulu, koko);  
  printf("Samoja lukuja peilikuvasijainneissa oli %" PRId64 " kpl.\n", samoja);
#endif

  return 0;
}
